# SPDX-FileCopyrightText: 2021-2024 The Ikarus Developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later

from dune.common.hashit import hashIt
from .generator import MySimpleGenerator

def dirichletValues(basis):
    """
    @brief Creates a Dirichlet values handler for the given basis.

    @param basis: The basis.

    @return: The created Dirichlet values handler.
    """
    generator = MySimpleGenerator("DirichletValues", "Ikarus::Python")
    element_type = f"Ikarus::DirichletValues<{basis.cppTypeName},Eigen::VectorX<bool>>"

    includes = []
    includes += basis._includes
    includes += ["ikarus/utils/dirichletvalues.hh"]
    includes += ["ikarus/python/dirichletvalues/dirichletvalues.hh"]
    moduleName = "dirichletValues_" + hashIt(element_type)
    module = generator.load(
        includes=includes, typeName=element_type, moduleName=moduleName
    )
    return module.DirichletValues(basis)



# SPDX-FileCopyrightText: 2021-2024 The Ikarus Developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later

import debug_info
debug_info.setDebugFlags()
import ikarus as iks
from ikarus import finite_elements, assembler
import numpy as np
import scipy as sp

from dune.iga import ControlPoint, ControlPointNet, NurbsPatchData, IGAGrid
from dune.iga.basis import defaultGlobalBasis, Power, Nurbs

from dune.common.hashit import hashIt
from dune.iga.basis import preBasisTypeName
from dune.generator.generator import SimpleGenerator


def globalBasis(gv, tree):
    generator = SimpleGenerator("BasisHandler", "Ikarus::Python")

    pbfName = preBasisTypeName(tree, gv.cppTypeName)
    element_type = f"Ikarus::BasisHandler<{pbfName}>"
    includes = []
    includes += list(gv.cppIncludes)
    includes += ["dune/iga/nurbsbasis.hh"]
    includes += ["ikarus/python/basis/basis.hh"]

    moduleName = "Basis_" + hashIt(element_type)
    module = generator.load(
        includes=includes, typeName=element_type, moduleName=moduleName
    )
    basis = defaultGlobalBasis(gv, tree)
    return module.BasisHandler(basis)


if __name__ == "__main__":
    cp = ControlPoint((0, 0, 0), 1)
    cp2 = ControlPoint((10, 0, 0), 1)
    cp3 = ControlPoint((0, 2, 0), 1)
    cp4 = ControlPoint((10, 2, 0), 1)

    netC = ((cp, cp2), (cp3, cp4))
    net = ControlPointNet(netC)

    nurbsPatchData1 = NurbsPatchData(((0, 0, 1, 1), (0, 0, 1, 1)), net, (1, 1))
    nurbsPatchData2 = nurbsPatchData1.degreeElevate(0, 1)
    nurbsPatchDataFinal = nurbsPatchData2.degreeElevate(1, 1)
    # nurbsPatchDataFinal = nurbsPatchData2.knotRefinement((0.1,0.2,0.3,0.4,0.5), 0)
    # # nurbsPatchData = nurbsPatchData.knotRefinement((0.25,0.5,0.75), 1)

    gridView = IGAGrid(nurbsPatchDataFinal)

    gridView.hierarchicalGrid.globalRefine(2)
    basis = globalBasis(gridView, Power(Nurbs(), 3))

    flatBasis = basis.flat()
    print("Number of dofs: ", len(flatBasis))
    d = np.zeros(len(flatBasis))

    lambdaLoad = iks.Scalar(1.0)
    thickness = 0.1

    def vL(x, lambdaVal):
        return np.array([0, 0, 2 * thickness**3 * lambdaVal])

    vLoad = iks.finite_elements.volumeLoad3D(vL)

    klShell = iks.finite_elements.kirchhoffLoveShell(youngs_modulus=1000, nu=0.0,thickness=thickness)

    fes = []
    for e in gridView.elements:
        fes.append(iks.finite_elements.makeFE(basis,klShell,vLoad))
        fes[-1].bind(e)

    dirichletValues = iks.dirichletValues(flatBasis)

    def fixLeftAndRightEdge(vec, localIndex, localView, intersection):
        if (
            intersection.geometry.center[0] < 1e-8
            or intersection.geometry.center[0] > 10 - 1e-8
        ):
            vec[localView.index(localIndex)] = True

    dirichletValues.fixBoundaryDOFsUsingLocalViewAndIntersection(fixLeftAndRightEdge)

    assembler = iks.assembler.sparseFlatAssembler(fes, dirichletValues)

    def gradAndhess(dRedInput):
        req = fes[0].createRequirement()
        req.insertParameter( lambdaLoad)
        dBig = assembler.createFullVector(dRedInput)
        req.insertGlobalSolution( dBig)
        g = assembler.vector(req,iks.VectorAffordance.forces, iks.DBCOption.Reduced)
        h = assembler.matrix(req,iks.MatrixAffordance.stiffness, iks.DBCOption.Reduced)
        return [g, h]

    from numpy.linalg import norm

    maxiter = 100
    abs_tolerance = 1e-8
    d = np.zeros(assembler.reducedSize())
    for k in range(maxiter):
        R, K = gradAndhess(d)
        r_norm = norm(R)

        deltad = sp.sparse.linalg.spsolve(K, R)
        d -= deltad
        print(k, r_norm, norm(deltad))
        if r_norm < abs_tolerance:
            break

    dBig = assembler.createFullVector(d)
    displacementFunc = flatBasis.asFunction(dBig)
    vtkWriter = gridView.trimmedVtkWriter()
    vtkWriter.addPointData(displacementFunc, name="displacement")
    vtkWriter.write("KLshell")
    assert (0.2087577577980777 - max(d)) < 1e-6, f"The maximum displacement should be 0.2087577577980777 but is {max(d)}"



# SPDX-FileCopyrightText: 2021-2024 The Ikarus Developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later

from dune.common.hashit import hashIt
from ikarus.generator import MySimpleGenerator

def sparseFlatAssembler(fes, dirichletValues):
    """
    @brief Creates a sparse flat assembler for the given finite elements and Dirichlet values.

    @param fes: The list of finite elements.
    @param dirichletValues: The Dirichlet values.

    @return: The created sparse flat assembler.
    """
    element_type = f"Ikarus::SparseFlatAssembler<std::vector<{fes[0].cppTypeName}>,{dirichletValues.cppTypeName}>"
    generator = MySimpleGenerator("SparseFlatAssembler", "Ikarus::Python")

    includes = []
    includes += ["ikarus/assembler/simpleassemblers.hh"]
    includes += fes[0].cppIncludes  # include header of finite element
    includes += dirichletValues.cppIncludes
    includes += ["ikarus/python/assembler/flatassembler.hh"]
    moduleName = "SparseFlatAssembler_" + hashIt(element_type)
    module = generator.load(
        includes=includes, typeName=element_type, moduleName=moduleName
    )
    return module.SparseFlatAssembler(fes, dirichletValues)


def denseFlatAssembler(fes, dirichletValues):
    """
    @brief Creates a dense flat assembler for the given finite elements and Dirichlet values.

    @param fes: The list of finite elements.
    @param dirichletValues: The Dirichlet values.

    @return: The created dense flat assembler.
    """
    element_type = f"Ikarus::DenseFlatAssembler<std::vector<{fes[0].cppTypeName}>,{dirichletValues.cppTypeName}>"
    generator = MySimpleGenerator("DenseFlatAssembler", "Ikarus::Python")

    includes = []
    includes += ["ikarus/assembler/simpleassemblers.hh"]
    includes += ["ikarus/python/assembler/flatassembler.hh"]
    includes += fes[0].cppIncludes  # include header of finite element
    includes += dirichletValues.cppIncludes
    moduleName = "DenseFlatAssembler_" + hashIt(element_type)
    module = generator.load(
        includes=includes, typeName=element_type, moduleName=moduleName
    )
    return module.DenseFlatAssembler(fes, dirichletValues)

def assemblerManipulator(assembler):
    """
    @brief Creates an assembler manipulator for the given assembler. The manipulation happens by providing call back functions

    @param assembler: The assembler.

    @return: The created assembler manipulator.
    """
    element_type = f"decltype(Ikarus::makeAssemblerManipulator(std::declval<{assembler.cppTypeName}>()))"
    generator = MySimpleGenerator("AssemblerManipulator", "Ikarus::Python")

    includes = []
    includes += assembler.cppIncludes
    includes += ["ikarus/assembler/assemblermanipulatorfuser.hh"]
    includes += ["ikarus/python/assembler/flatassemblermanipulator.hh"]
    moduleName = "AssemblerManipulator_" + hashIt(element_type)
    module = generator.load(
        includes=includes, typeName=element_type, moduleName=moduleName
    )
    return module.AssemblerManipulator(assembler)



# SPDX-FileCopyrightText: 2021-2024 The Ikarus Developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later
try:
    from dune.packagemetadata import registerExternalModule
    import pathlib

    # register ikarus to be recognized by dune-py (code generation module)
    # as a module of the dune universe
    registerExternalModule(
        moduleName="ikarus",
        modulePath=str(pathlib.Path(__file__).parent.resolve()),
    )

except ImportError:
    pass



import warnings
from ._ikarus import *
from .dirichlet_values import dirichletValues
from .basis import basis

from warnings import warn
def ValueWrapper(*args, **kwargs):
    warnings.warn("The 'ValueWrapper' class was renamed to Scalar. Please use 'Scalar' instead.",
                  DeprecationWarning, stacklevel=2 )
    return _ikarus.Scalar(*args, **kwargs)


# SPDX-FileCopyrightText: 2021-2024 The Ikarus Developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later
import os
os.environ["DUNE_LOG_LEVEL"] = "debug"
os.environ["DUNE_SAVE_BUILD"] = "terminal"

def setDebugFlags():
    import dune.generator as generator
    generator.setFlags("-g ",noChecks=False)

def unsetDebugFlags():
    import dune.generator as generator
    generator.reset()



# SPDX-FileCopyrightText: 2021-2024 The Ikarus Developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later

from dune.common.hashit import hashIt
from ikarus.generator import MySimpleGenerator

def registerPreElement(name, includes, element_type, *args):
    """
    @brief Registers a pre-element with the specified parameters.

    @param name: The name of the pre-element.
    @param includes: List of additional include files.
    @param element_type: The type of the pre-element.
    @param args: Additional arguments required for the pre-element registration.

    @return: The registered pre-element function.
    """
    generator = MySimpleGenerator(name, "Ikarus::Python")
    includes += ["ikarus/python/finiteelements/registerpreelement.hh"]

    moduleName = name + "_" + hashIt(element_type)
    module = generator.load(
        includes=includes, typeName=element_type, moduleName=moduleName
    )

    f = getattr(module, name)
    return f(*args)

def volumeLoad(f, d: int):
    """
    @brief Creates a volume load pre-element for the specified dimension.

    @param f: The volume load function.
    @param d: The dimension of the volume load (1, 2, or 3).

    @return: The registered volume load pre-element function.
    """
    includes = ["ikarus/finiteelements/mechanics/loads/volume.hh"]
    includes += ["dune/python/pybind11/functional.h"]
    includes += ["dune/python/pybind11/stl.h"]
    includes += ["dune/python/pybind11/eigen.h"]

    element_type = f"Ikarus::VolumeLoadPre<{d}>"
    return registerPreElement("VolumeLoadPre", includes, element_type, f)

def volumeLoad1D(f):
    """
    @brief Creates a 1D volume load pre-element.

    @param f: The volume load function.

    @return: The registered 1D volume load pre-element function.
    """
    return volumeLoad(f, 1)

def volumeLoad2D(f):
    """
    @brief Creates a 2D volume load pre-element.

    @param f: The volume load function.

    @return: The registered 2D volume load pre-element function.
    """
    return volumeLoad(f, 2)

def volumeLoad3D(f):
    """
    @brief Creates a 3D volume load pre-element.

    @param f: The volume load function.

    @return: The registered 3D volume load pre-element function.
    """
    return volumeLoad(f, 3)

def neumannBoundaryLoad(boundaryPatch, f):
    """
    @brief Creates a Neumann boundary load pre-element.

    @param boundaryPatch: The boundary patch.
    @param f: The traction function.

    @return: The registered Neumann boundary load pre-element function.
    """
    includes = ["ikarus/finiteelements/mechanics/loads/traction.hh"]
    includes += ["dune/python/pybind11/functional.h"]
    includes += ["dune/python/pybind11/stl.h"]
    includes += ["dune/python/pybind11/eigen.h"]
    includes += boundaryPatch._includes
    element_type = f"Ikarus::NeumannBoundaryLoadPre<{boundaryPatch.gridView().cppTypeName}>"
    return registerPreElement("NeumannBoundaryLoadPre", includes, element_type, boundaryPatch, f)

def nonLinearElastic(mat):
    """
    @brief Creates a non-linear elastic pre-element.

    @param mat: The material.

    @return: The registered non-linear elastic pre-element function.
    """
    includes = ["ikarus/finiteelements/mechanics/nonlinearelastic.hh"]

    element_type = f"Ikarus::NonLinearElasticPre<{mat.cppTypeName}>"
    includes += mat._includes
    return registerPreElement("NonLinearElasticPre", includes, element_type, mat)

def linearElastic(youngs_modulus, nu):
    """
    @brief Creates a linear elastic pre-element.

    @param youngs_modulus: Young's modulus.
    @param nu: Poisson's ratio.

    @return: The registered linear elastic pre-element function.
    """
    includes = ["ikarus/finiteelements/mechanics/linearelastic.hh"]
    element_type = "Ikarus::LinearElasticPre"
    return registerPreElement("LinearElasticPre", includes, element_type, youngs_modulus, nu)

def eas(numberofparameters):
    """
    @brief Creates an enhanced assumed strains pre-element.

    @param numberofparameters: The number of parameters.

    @return: The registered enhanced assumed strains pre-element function.
    """
    includes = ["ikarus/finiteelements/mechanics/enhancedassumedstrains.hh"]
    element_type = "Ikarus::EnhancedAssumedStrainsPre"
    return registerPreElement("EnhancedAssumedStrainsPre", includes, element_type, numberofparameters)

def kirchhoffLoveShell(youngs_modulus: float, nu, thickness):
    """
    @brief Creates a Kirchhoff Love shell pre-element.

    @param youngs_modulus: Young's modulus.
    @param nu: Poisson's ratio.
    @param thickness: The thickness of the shell.

    @return: The registered Kirchhoff Love shell pre-element function.
    """
    includes = ["ikarus/finiteelements/mechanics/kirchhoffloveshell.hh"]
    element_type = "Ikarus::KirchhoffLoveShellPre"
    return registerPreElement("KirchhoffLoveShellPre", includes, element_type, youngs_modulus, nu, thickness)

def makeFE(basis, *skills):
    """
    @brief Creates a finite element with the specified basis and skills.

    @param basis: The basis for the finite element.
    @param skills: Skills for the finite element.

    @return: The created finite element.
    """
    includes = ["ikarus/python/finiteelements/fe.hh"]
    preFE_type = f"Ikarus::PreFE<{basis.cppTypeName},true,true>"
    element_type = f"Ikarus::FE<{preFE_type},"

    for arg in skills:
        element_type += arg.cppTypeName + "::Skill,"
    element_type = element_type[:-1]
    element_type += ">"

    generator = MySimpleGenerator("FE", "Ikarus::Python")

    includes += basis._includes
    for arg in skills:
        includes += arg._includes

    moduleName = "FE" + "_" + hashIt(element_type)
    module = generator.load(
        includes=includes, typeName=element_type, moduleName=moduleName
    )

    tupleskills = tuple(skills)
    return module.FE(basis, tupleskills)



# SPDX-FileCopyrightText: 2021-2024 The Ikarus Developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later

import os
import re
import datetime


def update_license_header(file_path):
    try:
        with open(file_path, "r", newline="") as file:
            content = file.read()

        # Update the license header using regex
        updated_content, nsubs = re.subn(
            r"SPDX-FileCopyrightText: (\d{4}-\d{4}) The Ikarus Developers",
            lambda match: f"SPDX-FileCopyrightText: 2021-{datetime.datetime.now().year} The Ikarus Developers",
            content,
        )
        if nsubs > 0:
            with open(file_path, "w", newline="") as file:
                file.write(updated_content)
            print(f"Updated license header in: {file_path}")
        else:
            print(f"Didn't update license header in: {file_path}")

    except Exception as e:
        print(f"Error updating license header in {file_path}: {e}")


def update_dep5(file_path):
    try:
        with open(file_path, "r", newline="") as file:
            content = file.read()

        updated_content, nsubs = re.subn(
            r"Copyright: (\d{4}-\d{4}) The Ikarus Developers",
            lambda match: f"Copyright: 2021-{datetime.datetime.now().year} The Ikarus Developers",
            content,
        )
        if nsubs > 0:
            with open(file_path, "w", newline="") as file:
                file.write(updated_content)

    except Exception as e:
        print(f"Error updating license information in {file_path}: {e}")


def process_directory(directory):
    for root, _, files in os.walk(directory):
        for file in files:
            file_path = os.path.join(root, file)
            if os.path.basename(file_path) == 'dep5':
                update_dep5(file_path)
            else:
                update_license_header(file_path)


if __name__ == "__main__":
    script_dir = os.path.dirname(os.path.realpath("__file__"))

    print(script_dir)
    process_directory(script_dir)
